"""
Tamarin is a Django app for retrieving, parsing, and storing S3 logs.

Interesting bits are in models.py (Django model definitions), log_puller.py
(the functions used to set all of the pieces in motion), and parser.py (the
PyParsing-based log parser). The rest are just your typical support modules
seen in most Django apps.
"""
# Major, minor.
VERSION = (1, 0)



"""
ModelAdmin classes for the tamarin package.
"""
from django.contrib import admin
from tamarin.models import S3LoggedBucket, S3LogRecord

class S3LoggedBucketAdmin(admin.ModelAdmin):
    """
    Admin configuration for S3LoggedBucket model.
    """
    list_display = ['name', 'log_bucket_name', 'monitor_bucket']
    list_filter = ['monitor_bucket']
    search_fields = ['name', 'log_bucket_name']
    ordering = ['name']
admin.site.register(S3LoggedBucket, S3LoggedBucketAdmin)

class S3LogRecordAdmin(admin.ModelAdmin):
    """
    Admin configuration for S3LogRecord model.
    """
    list_display = [
        'admin_bucket_name', 'key', 'remote_ip', 'http_status',
        'bytes_sent', 'object_size', 'request_dtime',
    ]
    list_filter = ['http_status']
    search_fields = ['request_ip', 'request_uri']
    ordering = ['-request_dtime']

    def admin_bucket_name(self, obj):
        """
        Returns the name of the log's monitored bucket.
        
        :param S3LogRecord obj: The object whose bucket name to retrieve.
        :rtype: str
        :returns: The name of the bucket being monitored in this log.
        """
        return obj.bucket.name

admin.site.register(S3LogRecord, S3LogRecordAdmin)



"""
Django models for storing S3 log data. The main event is :class:`S3LogRecord`.
"""
from django.db import models

class S3LoggedBucketManager(models.Manager):
    """
    Table-level operations for S3LoggedBucket model.
    """
    def get_log_buckets_to_monitor(self):
        """
        Returns which S3LoggedBucket objects the log puller should monitor
        for log keys.
        
        :rtype: QuerySet
        :returns: The S3LoggedBucket objects that should be monitored for
            log keys.
        """
        return self.filter(monitor_bucket=True)

class S3LoggedBucket(models.Model):
    """
    A bucket that is logged.
    """
    name = models.CharField(max_length=255, unique=True, db_index=True,
        help_text="The name of the bucket that is being logged. This is "\
                  "where the media that is being served resides.")
    log_bucket_name = models.CharField(max_length=255,
        help_text="The S3 bucket to monitor for log keys. This must be a "\
                  "separate bucket from the logged bucket. The contents of "\
                  "this bucket must only be log files.")
    monitor_bucket = models.BooleanField(default=True,
        help_text="When checked, pull logs from this bucket's log bucket.")

    # Custom manager.
    objects = S3LoggedBucketManager()

    def __unicode__(self):
        return "S3LoggedBucket: %s" % self.name

class S3LogRecord(models.Model):
    """
    A single log entry for a bucket in S3.
    """
    bucket_owner = models.CharField(max_length=255,
        help_text="The canonical id of the owner of the source bucket.")
    bucket = models.ForeignKey(S3LoggedBucket,
        help_text="The bucket that the request was processed against.")
    request_dtime = models.DateTimeField(
        help_text="The time at which the request was received.")
    remote_ip = models.IPAddressField(
        help_text="The apparent Internet address of the requester. "\
                  "Intermediate proxies and firewalls might obscure the "\
                  "actual address of the machine making the request.")
    requester = models.CharField(max_length=255, blank=True, null=True,
        help_text="The canonical user id of the requester, or the string "\
                  "'Anonymous' for unauthenticated requests. This "\
                  "identifier is the same one used for access control "\
                  "purposes.")
    request_id = models.CharField(max_length=255, unique=True, db_index=True,
        help_text="A string generated by Amazon S3 to uniquely identify "\
                  "each request.")
    operation = models.CharField(max_length=255,
        help_text="Either SOAP.<operation> or "\
                  "REST.<HTTP_method.resource_type>.")
    key = models.TextField(
        help_text="The 'key' part of the request, URL encoded, or '-' if "\
                  "the operation does not take a key parameter.")
    request_method = models.CharField(max_length=10, blank=True, null=True,
        help_text="The method used to retrieve the file. Typically either "\
                  "GET or POST.")
    request_uri = models.TextField(
        help_text="The Request-URI part of the HTTP request message.")
    http_version = models.CharField(max_length=15, blank=True, null=True,
        help_text="HTTP version used in the request. Typically HTTP/1.0 "\
                  "or HTTP/1.1.")
    http_status = models.PositiveIntegerField(
        help_text="The HTTP response code for the request.")
    error_code = models.CharField(max_length=255, blank=True, null=True,
        help_text="The Amazon S3 Error Code, or '-' if no error occurred.")
    bytes_sent = models.PositiveIntegerField(blank=True, null=True,
        help_text="The number of response bytes sent, excluding HTTP "\
                  "protocol overhead, or '-' if zero.")
    object_size = models.PositiveIntegerField(blank=True, null=True,
        help_text="The total size of the object in question.")
    total_time = models.PositiveIntegerField(blank=True, null=True,
        help_text="The number of milliseconds the request was in flight "\
                  "from the server's perspective. This value is measured "\
                  "from the time your request is received to the time that "\
                  "the last byte of the response is sent. Measurements made "\
                  "from the client's perspective might be longer due to "\
                  "network latency.")
    turnaround_time = models.PositiveIntegerField(blank=True, null=True,
        help_text="The number of milliseconds that Amazon S3 spent "\
                  "processing your request. This value is measured from "\
                  "the time the last byte of your request was received until "\
                  "the time the first byte of the response was sent.")
    referrer = models.TextField(blank=True, null=True,
        help_text="The value of the HTTP Referrer header, if present. "\
                  "HTTP user-agents (e.g. browsers) typically set this "\
                  "header to the URL of the linking or embedding page when "\
                  "making a request.")
    user_agent = models.TextField(blank=True, null=True,
        help_text="The value of the HTTP User-Agent header.")
    version_id = models.CharField(max_length=255, blank=True, null=True,
        help_text="The version ID in the request, or '-' if the operation "\
                  "does not take a versionId parameter.")

    # The following fields (by string name) have dash values replaced with
    # None values so we don't have to query by '-'.
    DASHED_TO_NONE_FIELDS = [
        'requester',
        'error_code',
        'referrer',
        'object_size',
        'version_id',
        'user_agent',
        'turnaround_time',
        'object_size',
        'bytes_sent',
        'total_time',
        'user_agent',
    ]

    def __unicode__(self):
        return "S3LogRecord: %s" % self.id

    def save(self):
        """
        Look through the dash-to-empty fields, replace dashes with a 
        None value to promote less silly query filters.
        """
        for field in self.DASHED_TO_NONE_FIELDS:
            field_val = getattr(self, field)
            if field_val == '-' or field_val == '':
                setattr(self, field, None)

        super(S3LogRecord, self).save()



# encoding: utf-8
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models

class Migration(SchemaMigration):

    def forwards(self, orm):
        
        # Changing field 'S3LogRecord.bytes_sent'
        db.alter_column('tamarin_s3logrecord', 'bytes_sent', self.gf('django.db.models.fields.PositiveIntegerField')(null=True))


    def backwards(self, orm):
        
        # User chose to not deal with backwards NULL issues for 'S3LogRecord.bytes_sent'
        raise RuntimeError("Cannot reverse this migration. 'S3LogRecord.bytes_sent' and its values cannot be restored.")


    models = {
        'tamarin.s3loggedbucket': {
            'Meta': {'object_name': 'S3LoggedBucket'},
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'log_bucket_name': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            'monitor_bucket': ('django.db.models.fields.BooleanField', [], {'default': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '255', 'db_index': 'True'})
        },
        'tamarin.s3logrecord': {
            'Meta': {'object_name': 'S3LogRecord'},
            'bucket': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['tamarin.S3LoggedBucket']"}),
            'bucket_owner': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            'bytes_sent': ('django.db.models.fields.PositiveIntegerField', [], {'null': 'True', 'blank': 'True'}),
            'error_code': ('django.db.models.fields.CharField', [], {'max_length': '255', 'blank': 'True'}),
            'http_status': ('django.db.models.fields.PositiveIntegerField', [], {}),
            'http_version': ('django.db.models.fields.CharField', [], {'max_length': '15'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'key': ('django.db.models.fields.TextField', [], {}),
            'object_size': ('django.db.models.fields.PositiveIntegerField', [], {'null': 'True', 'blank': 'True'}),
            'operation': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            'referrer': ('django.db.models.fields.TextField', [], {'blank': 'True'}),
            'remote_ip': ('django.db.models.fields.IPAddressField', [], {'max_length': '15'}),
            'request_dtime': ('django.db.models.fields.DateTimeField', [], {}),
            'request_id': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '255', 'db_index': 'True'}),
            'request_method': ('django.db.models.fields.CharField', [], {'max_length': '10'}),
            'request_uri': ('django.db.models.fields.TextField', [], {}),
            'requester': ('django.db.models.fields.CharField', [], {'max_length': '255', 'blank': 'True'}),
            'total_time': ('django.db.models.fields.PositiveIntegerField', [], {}),
            'turnaround_time': ('django.db.models.fields.PositiveIntegerField', [], {'null': 'True', 'blank': 'True'}),
            'user_agent': ('django.db.models.fields.TextField', [], {}),
            'version_id': ('django.db.models.fields.CharField', [], {'max_length': '255', 'blank': 'True'})
        }
    }

    complete_apps = ['tamarin']



"""
The tamarin_pull_logs management command may be used in conjunction with
cron or another task scheduler if you don't have or want to use celery.
"""
from django.core.management.base import BaseCommand
from tamarin.log_puller import pull_and_parse_logs

class Command(BaseCommand):
    """
    Nothing magical here, just using the log puller.
    """
    help = "Looks through S3 logs for broadcast media to find the hogs."

    def handle(self, *args, **options):
        """
        Call the log pulling function and do work silently.
        """
        pull_and_parse_logs()



"""
This module is used to put all of the pieces together and provide a single
point of entry to get logs pulled, parsed, and stored. The only important
function in this module for users is :func:`pull_and_parse_logs`. This is what
the celery task and the Django management command (tamarin_pull_logs) call
to get things going.
"""
import logging
import boto
from boto.exception import S3ResponseError
from django.conf import settings
from tamarin.models import S3LoggedBucket, S3LogRecord
from tamarin.parser import S3LogParser

# The following fields should not be pulled from the parser's results. They
# are either set when the model is saved (id), or in store_parsed_entry.
MANUAL_FIELDS = ['id', 'bucket']
# Logger for this module.
LOGGER = logging.getLogger(__name__)

def store_parsed_entry(parsed):
    """
    Given a parsed log entry, copy the values from the parser's output to
    a S3LogRecord object. If the log already exists, update the values.
    
    :param pyparsing.ParseResults parsed: The parsed results for a single
        log entry.
    :rtype: S3LogRecord
    :returns: The updated or newly created S3LogRecord object for the
        parsed log entry.
    """
    try:
        record = S3LogRecord.objects.get(request_id=parsed['request_id'])
    except S3LogRecord.DoesNotExist:
        record = S3LogRecord()

    field_names = [field.name for field in record._meta.fields \
                      if field.name not in MANUAL_FIELDS]
    for field in field_names:
        setattr(record, field, parsed.get(field))

    record.bucket = S3LoggedBucket.objects.get(name=parsed['bucket'])
    record.save()
    return record

def pull_and_parse_logs():
    """
    Pulls all of the keys from the bucket's log bucket, hands them off to the
    parser, then stores the values in S3LogRecord objects.
    """
    # When True, delete S3 log keys after they have been parsed.
    purge_parsed_keys = getattr(settings, 'TAMARIN_PURGE_PARSED_KEYS', True)
    # All buckets with monitor_bucket == True (active).
    logged_buckets = S3LoggedBucket.objects.get_log_buckets_to_monitor()
    conn = boto.connect_s3(settings.AWS_ACCESS_KEY_ID,
                           settings.AWS_SECRET_ACCESS_KEY)

    for logged_bucket in logged_buckets:
        bucket = conn.get_bucket(logged_bucket.log_bucket_name)
        for key in bucket.get_all_keys():
            LOGGER.debug(key)

            try:
                # Stuff the entire contents of the S3 key into string variable.
                log_contents = key.get_contents_as_string()
            except S3ResponseError:
                # We're going to get a response error from time to time, S3
                # is not bullet proof. In this case, move on to the next
                # key and a later execution of the puller will grab this one.
                print "S3ResponseError encountered when retrieving: %s" % key
                continue

            parser = S3LogParser(log_contents)
            for log_entry in parser.parse_lines():
                store_parsed_entry(log_entry)

            if purge_parsed_keys:
                # Delete the key after parsing it if the
                # TAMARIN_PURGE_PARSED_KEYS setting is True. 
                key.delete()



from setuptools import setup
import tamarin

DESCRIPTION = "A Django app for monitoring AWS usage in Django's admin."

LONG_DESCRIPTION = None
try:
    LONG_DESCRIPTION = open('README.rst').read()
except:
    pass

version_str = '%d.%d' % (tamarin.VERSION[0], tamarin.VERSION[1])

CLASSIFIERS = [
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: BSD License',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Topic :: Software Development :: Libraries :: Python Modules',
    'Framework :: Django',
]

setup(
    name='tamarin',
    version=version_str,
    packages=[
        'tamarin',
        'tamarin.management', 'tamarin.management.commands',
        'tamarin.migrations',
    ],
    author='Gregory Taylor',
    author_email='gtaylor@duointeractive.com',
    url='https://github.com/duointeractive/tamarin/',
    license='BSD',
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    platforms=['any'],
    classifiers=CLASSIFIERS,
    install_requires=['boto', 'pyparsing'],
)



"""
S3 log file parser, using PyParsing. The S3LogParser is instantiated with
the full string contents of a log key, and delegates the parsing of each line
to the S3LogLineParser class.

The primary user of this module is the `log_puller` module.
"""
import logging
import datetime
from dateutil import zoneinfo
from pyparsing import alphas, nums, alphanums, Combine, Word, Group, delimitedList, Suppress
from django.conf import settings

LOGGER = logging.getLogger(__name__)

class S3LogLineParser(object):
    """
    Parses individual lines within an S3 bucket log. Spits out 
    S3LogRecord objects.
    """
    FIELDS = (
        'bucket', 'request_dtime', 'remote_ip', 'requester',
        'request_id', 'operation', 'key', 'request_method',
        'request_uri', 'http_version', 'http_status', 'error_code',
        'bytes_sent', 'object_size', 'total_time', 'turnaround_time',
        'referrer', 'user_agent', 'version_id',
    )

    def __init__(self, line_contents):
        """
        :param str line_contents: The individual line/log entry to parse, in
            string form.
        """
        self.line_contents = line_contents

    def parse_line(self):
        """
        Parses the line that the S3LogLineParser class was instantiated with.
        
        :rtype: pyparsing.ParseResults
        :returns: A pyparsing ParseResults instance, which can be used much
            like a dict.
        """
        LOGGER.debug("-- RAW --" * 5)
        LOGGER.debug(self.line_contents)
        LOGGER.debug("-- RESULTS --" * 5)
        results = self.parse()
        LOGGER.debug(self._parse_result_debug_msg(results))
        return results

    def _action_dtime_parse(self, full_str, length, t):
        """
        Parses an S3-formatted time string into a datetime.datetime instance.
        
        :param str full_str: The full, un-parsed time string.
        :param int length: The length of the text to parse.
        :param list t: The grouped string to parse.
        :rtype: datetime.datetime
        :returns: The datetime.datetime equivalent of the time string with
            the correct (settings.TIME_ZONE) timezone set.
        """
        # Comes in as [['22/Apr/2011:18:28:10', '+000']] but we're just
        # interested in the time, since strptime with a %z formatter doesn't
        # work as expected on all platforms.
        dtime_str = t[0][0]
        # 22/Apr/2011:18:28:10
        utc = datetime.datetime.strptime(dtime_str, "%d/%b/%Y:%H:%M:%S")
        # The parsed time is in UTC. Make it "aware".
        utc = utc.replace(tzinfo=zoneinfo.gettz('UTC'))
        if settings.TIME_ZONE != 'UTC':
            # Get the user's local timezone.
            to_zone = zoneinfo.gettz(settings.TIME_ZONE)
            # Set the timezone to the configured TZ.
            return utc.astimezone(to_zone)
        else:
            # Already UTC, don't budge.
            return utc

    def parse(self):
        """
        Parses the class's :attr:`line_contents` attribute using pyparsing.
        
        :rtype: pyparsing.ParseResults
        :returns: A pyparsing ParseResults instance, which can be used much
            like a dict.
        """
        # Some generic parsing patterns.
        integer = Word(nums)
        generic_alphanum = Word(alphanums)
        alpha_or_dash = Word(alphas + "-")
        num_or_dash = Word(nums + '-')

        # S3 bucket key name. Restrictions apply.
        bucket_str = Word(alphanums + '-_.')
        # IE: +400 or -400
        time_zone_offset = Word("+-", nums)
        # Abbreviated month: Jan, Feb, etc.
        month = Word(alphas, exact=3)
        # [04/Aug/2006:22:34:02 +0000]
        server_dtime = Group(
                             Suppress("[") +
                             Combine(
                                 integer + "/" +
                                 month + "/" +
                                 integer + ":" +
                                 integer + ":" +
                                 integer + ":" +
                                 integer) +
                             time_zone_offset +
                             Suppress("]")
                         )
        # 72.21.206.5
        ip_address = delimitedList(integer, ".", combine=True)
        # 314159b66967d86f031c7249d1d9a80 or -
        requester = Word(alphanums + "-")
        # IE: SOAP.CreateBucket or REST.PUT.OBJECT
        operation = Word(alphas + "._")
        # S3 key: /photos/2006/08/puppy.jpg
        key = Word(alphanums + "/-_.?=%&:+<>#~[]")
        # One of GET, POST, or PUT
        http_method = Word(alphas)
        # HTTP/1.1
        http_protocol = Word(alphanums + "/.")

        # "GET /mybucket/photos/2006/08/ HTTP/1.1"
        uri = Suppress('"') + \
              http_method('request_method') + \
              key("request_uri") + \
              http_protocol('http_version') + \
              Suppress('"')
        dash_or_uri = ("-" | uri)

        # "http://www.amazon.com/webservices"
        referrer_uri = Suppress('"') + key + Suppress('"')
        # Referrer can be empty double quotes sometimes.
        empty_dquotes = Suppress('"') + Suppress('"')
        # Either a referrer, a dash, or empty double quotes.
        referrer_or_dash = referrer_uri | "-" | empty_dquotes

        # "curl/7.15.1"
        user_agent = Suppress('"') + \
                     Word(alphanums + "/-_.?=%&:(); ,+$@!^<>~[]'{}#*`") + \
                     Suppress('"')
        # User agent field can either be a user agent string or a dash.
        user_agent_or_dash = user_agent | "-"

        # The string value for each field below is what you refer to when
        # accessing the parsed values.
        log_line_bnf = (
            generic_alphanum("bucket_owner") +
            bucket_str("bucket") +
            server_dtime("request_dtime").setParseAction(self._action_dtime_parse) +
            ip_address("remote_ip") +
            requester("requester") +
            generic_alphanum("request_id") +
            operation("operation") +
            key("key") +
            dash_or_uri("request_uri") +
            integer('http_status') +
            alpha_or_dash('error_code') +
            num_or_dash('bytes_sent') +
            num_or_dash('object_size') +
            num_or_dash('total_time') +
            num_or_dash('turnaround_time') +
            referrer_or_dash('referrer') +
            user_agent_or_dash('user_agent') +
            alpha_or_dash('version_id')
        )
        return log_line_bnf.parseString(self.line_contents)

    def _parse_result_debug_msg(self, parsed):
        """
        Used to show useful output during debugging. Shows all of the parsed
        fields and their values.
        
        :param pyparsing.ParseResults parsed: A parsed log line.
        :rtype: str
        :return: A debug message that may be sent through the logger.
        """
        retval = ""
        for field in self.FIELDS:
            retval += "%s: %s\n" % (field, parsed.get(field, ''))
        return retval

class S3LogParser(object):
    """
    Handles the parsing of S3 bucket log key contents.
    """
    def __init__(self, log_contents):
        """
        :param str log_contents: The contents of the log file, in string form.
        """
        self.log_contents = log_contents

    def parse_lines(self):
        """
        Breaks the file up into lines and delegates the parsing of each
        line to :class:`S3LogLineParser`.
        
        :rtype: pyparsing.ParseResults
        :returns: A pyparsing ParseResults instance, which can be used much
            like a dict.
        """
        lines = self.log_contents.split('\n')
        for line in lines:
            if not line:
                # Empty line, skip.
                continue
            # Delegate to the line parser.
            line_parser = S3LogLineParser(line)
            # Return the pyparsing.ParseResults for the line.
            yield line_parser.parse_line()



import datetime
from django.conf import settings
from celery.task import PeriodicTask
from tamarin.log_puller import pull_and_parse_logs

class PullAndParseS3Logs(PeriodicTask):
    """
    Pulls and parses S3 bucket logs at intervals. The
    TAMARIN_CELERY_PULL_PARSE_INTERVAL setting is used to determine the
    firing interval for this task. This number should decrease towards one
    minute as more buckets are monitored to prevent backlog.
    """
    pull_interval = getattr(settings, 'TAMARIN_CELERY_PULL_PARSE_INTERVAL', 5)
    run_every = datetime.timedelta(minutes=pull_interval)

    def run(self, **kwargs):
        """
        Do the same thing that the tamarin_pull_logs management command does.
        """
        pull_and_parse_logs()

